/*
 * Copyright 2015 - 2022 gary@drinkingtea.net
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at https://mozilla.org/MPL/2.0/.
 */

#pragma once

#include "strops.hpp"
#include "types.hpp"

namespace ox {

template<std::size_t buffLen>
class BString;

template<std::size_t buffLen>
class BasicString;

class StringView {
	private:
		const char *m_str = nullptr;
		std::size_t m_len = 0;
	public:
		constexpr StringView() noexcept = default;

		constexpr StringView(const StringView &sv) noexcept: m_str(sv.m_str), m_len(sv.m_len) {
		}

		template<std::size_t SmallStrSz>
		constexpr StringView(const BasicString<SmallStrSz> &str) noexcept: m_str(str.c_str()), m_len(str.len()) {}

		template<std::size_t SmallStrSz>
		constexpr StringView(const BString<SmallStrSz> &str) noexcept: m_str(str.c_str()), m_len(str.len()) {}

		constexpr StringView(std::nullptr_t) noexcept {}

		constexpr StringView(const char *str) noexcept: m_str(str), m_len(str ? ox_strlen(str) : 0) {}

		constexpr StringView(const char *str, std::size_t len) noexcept: m_str(str), m_len(len) {}

		[[nodiscard]]
		constexpr auto bytes() const noexcept {
			return m_len;
		}

		[[nodiscard]]
		constexpr auto len() const noexcept {
			return m_len;
		}

		[[nodiscard]]
		constexpr auto *data() const noexcept {
			return &m_str[0];
		}

		[[nodiscard]]
		constexpr auto &front() const noexcept {
			return m_str[0];
		}

		[[nodiscard]]
		constexpr auto &back() const noexcept {
			return m_str[m_len - 1];
		}

		constexpr auto operator[](std::size_t i) const noexcept {
			return m_str[i];
		}

		constexpr auto operator=(const StringView &other) noexcept {
			m_str = other.m_str;
			m_len = other.m_len;
			return *this;
		}

		constexpr auto operator==(const StringView &other) const noexcept {
			if (other.len() != len()) {
				return false;
			}
			return ox_strncmp(m_str, other.m_str, len()) == 0;
		}

		constexpr auto operator<=>(const StringView &str2) const noexcept {
			const StringView &str1 = *this;
			const auto maxLen = ox::min(str1.len(), str2.len());
			const auto a = &str1.front();
			const auto b = &str2.front();
			for (std::size_t i = 0; i < maxLen && (a[i] || b[i]); ++i) {
				if (a[i] < b[i]) {
					return -1;
				} else if (a[i] > b[i]) {
					return 1;
				}
			}
			if (str1.len() > str2.len()) {
				return 1;
			} else if (str1.len() < str2.len()) {
				return -1;
			} else {
				return 0;
			}
		}

};

using CRStringView = const StringView&;

static_assert(StringView("Read").bytes() == 4);
static_assert(StringView("Read") == StringView("Read"));
static_assert(StringView("Read") != StringView("Write"));
static_assert(StringView("Write") != StringView("Read"));
static_assert(StringView("Write") != StringView(""));
static_assert(StringView("") != StringView("Read"));
static_assert(StringView("") == StringView(""));
static_assert(StringView("") == "");
static_assert("Read" == StringView("Read"));
static_assert(StringView("Read") == StringView("Read"));

}

